@@ -45,6 +45,28 @@ impl Privkey {
         Ok(Privkey { obj })
     }
 
+    /// Create a new ElGamal private key with a random group
+    pub fn create_elgamal(
+        p_bits: usize,
+        q_bits: usize,
+        rng: &mut RandomNumberGenerator,
+    ) -> Result<Self> {
+        let obj = botan_init!(botan_privkey_create_elgamal, rng.handle(), p_bits, q_bits)?;
+
+        Ok(Self { obj })
+    }
+
+    /// Create a new DSA private key with a random group
+    pub fn create_dsa(
+        p_bits: usize,
+        q_bits: usize,
+        rng: &mut RandomNumberGenerator,
+    ) -> Result<Self> {
+        let obj = botan_init!(botan_privkey_create_dsa, rng.handle(), p_bits, q_bits)?;
+
+        Ok(Self { obj })
+    }
+
     /// Load an RSA private key (p,q,e)
     ///
     /// # Examples
@@ -219,10 +241,36 @@ impl Privkey {
 
     /// DER encode the key (unencrypted)
     pub fn der_encode(&self) -> Result<Vec<u8>> {
-        let der_len = 4096; // fixme
-        call_botan_ffi_returning_vec_u8(der_len, &|out_buf, out_len| unsafe {
-            botan_privkey_export(self.obj, out_buf, out_len, 0u32)
-        })
+        #[cfg(feature = "botan3")]
+        {
+            call_botan_ffi_viewing_vec_u8(&|ctx, cb| unsafe {
+                botan_privkey_view_der(self.obj, ctx, cb)
+            })
+        }
+
+        #[cfg(not(feature = "botan3"))]
+        {
+            call_botan_ffi_returning_vec_u8(4096, &|out_buf, out_len| unsafe {
+                botan_privkey_export(self.obj, out_buf, out_len, 0u32)
+            })
+        }
+    }
+
+    /// PEM encode the private key (unencrypted)
+    pub fn pem_encode(&self) -> Result<String> {
+        #[cfg(feature = "botan3")]
+        {
+            call_botan_ffi_viewing_str_fn(&|ctx, cb| unsafe {
+                botan_privkey_view_pem(self.obj, ctx, cb)
+            })
+        }
+
+        #[cfg(not(feature = "botan3"))]
+        {
+            call_botan_ffi_returning_string(4096, &|out_buf, out_len| unsafe {
+                botan_privkey_export(self.obj, out_buf, out_len, 1u32)
+            })
+        }
     }
 
     /// DER encode the key (encrypted)
@@ -250,24 +298,44 @@ impl Privkey {
         pbkdf_iter: usize,
         rng: &mut RandomNumberGenerator,
     ) -> Result<Vec<u8>> {
-        let der_len = 4096; // fixme
         let passphrase = make_cstr(passphrase)?;
         let cipher = make_cstr(cipher)?;
         let pbkdf = make_cstr(pbkdf)?;
 
-        call_botan_ffi_returning_vec_u8(der_len, &|out_buf, out_len| unsafe {
-            botan_privkey_export_encrypted_pbkdf_iter(
-                self.obj,
-                out_buf,
-                out_len,
-                rng.handle(),
-                passphrase.as_ptr(),
-                pbkdf_iter,
-                cipher.as_ptr(),
-                pbkdf.as_ptr(),
-                0u32,
-            )
-        })
+        let rng_handle = rng.handle();
+
+        #[cfg(feature = "botan3")]
+        {
+            call_botan_ffi_viewing_vec_u8(&|ctx, cb| unsafe {
+                botan_privkey_view_encrypted_der(
+                    self.obj,
+                    rng_handle,
+                    passphrase.as_ptr(),
+                    cipher.as_ptr(),
+                    pbkdf.as_ptr(),
+                    pbkdf_iter,
+                    ctx,
+                    cb,
+                )
+            })
+        }
+
+        #[cfg(not(feature = "botan3"))]
+        {
+            call_botan_ffi_returning_vec_u8(4096, &|out_buf, out_len| unsafe {
+                botan_privkey_export_encrypted_pbkdf_iter(
+                    self.obj,
+                    out_buf,
+                    out_len,
+                    rng_handle,
+                    passphrase.as_ptr(),
+                    pbkdf_iter,
+                    cipher.as_ptr(),
+                    pbkdf.as_ptr(),
+                    0u32,
+                )
+            })
+        }
     }
 
     /// PEM encode the key (encrypted)
@@ -295,41 +363,61 @@ impl Privkey {
         pbkdf_iter: usize,
         rng: &mut RandomNumberGenerator,
     ) -> Result<String> {
-        let pem_len = 4096; // fixme
-
         let passphrase = make_cstr(passphrase)?;
         let cipher = make_cstr(cipher)?;
         let pbkdf = make_cstr(pbkdf)?;
+        let rng_handle = rng.handle();
 
-        call_botan_ffi_returning_string(pem_len, &|out_buf, out_len| unsafe {
-            botan_privkey_export_encrypted_pbkdf_iter(
-                self.obj,
-                out_buf,
-                out_len,
-                rng.handle(),
-                passphrase.as_ptr(),
-                pbkdf_iter,
-                cipher.as_ptr(),
-                pbkdf.as_ptr(),
-                1u32,
-            )
-        })
-    }
+        #[cfg(feature = "botan3")]
+        {
+            call_botan_ffi_viewing_str_fn(&|ctx, cb| unsafe {
+                botan_privkey_view_encrypted_pem(
+                    self.obj,
+                    rng_handle,
+                    passphrase.as_ptr(),
+                    cipher.as_ptr(),
+                    pbkdf.as_ptr(),
+                    pbkdf_iter,
+                    ctx,
+                    cb,
+                )
+            })
+        }
 
-    /// PEM encode the private key (unencrypted)
-    pub fn pem_encode(&self) -> Result<String> {
-        let pem_len = 4096; // fixme
-        call_botan_ffi_returning_string(pem_len, &|out_buf, out_len| unsafe {
-            botan_privkey_export(self.obj, out_buf, out_len, 1u32)
-        })
+        #[cfg(not(feature = "botan3"))]
+        {
+            call_botan_ffi_returning_string(4096, &|out_buf, out_len| unsafe {
+                botan_privkey_export_encrypted_pbkdf_iter(
+                    self.obj,
+                    out_buf,
+                    out_len,
+                    rng_handle,
+                    passphrase.as_ptr(),
+                    pbkdf_iter,
+                    cipher.as_ptr(),
+                    pbkdf.as_ptr(),
+                    1u32,
+                )
+            })
+        }
     }
 
     /// Return the key agrement key, only valid for DH/ECDH
     pub fn key_agreement_key(&self) -> Result<Vec<u8>> {
-        let ka_key_len = 512; // fixme
-        call_botan_ffi_returning_vec_u8(ka_key_len, &|out_buf, out_len| unsafe {
-            botan_pk_op_key_agreement_export_public(self.obj, out_buf, out_len)
-        })
+        #[cfg(feature = "botan3")]
+        {
+            call_botan_ffi_viewing_vec_u8(&|ctx, cb| unsafe {
+                botan_pk_op_key_agreement_view_public(self.obj, ctx, cb)
+            })
+        }
+
+        #[cfg(not(feature = "botan3"))]
+        {
+            let ka_key_len = 512; // fixme
+            call_botan_ffi_returning_vec_u8(ka_key_len, &|out_buf, out_len| unsafe {
+                botan_pk_op_key_agreement_export_public(self.obj, out_buf, out_len)
+            })
+        }
     }
 
     /// Get a value for the private key
@@ -514,27 +602,57 @@ impl Pubkey {
     }
 
     /// Return hash of the public key data
-    pub fn fingerprint(&self, hash: &str) -> Result<String> {
+    pub fn fingerprint(&self, hash: &str) -> Result<Vec<u8>> {
         let hash = make_cstr(hash)?;
         let fprint_len = 64; // hashes > 512 bits are rare
-        call_botan_ffi_returning_string(fprint_len, &|out_buf, out_len| unsafe {
+        call_botan_ffi_returning_vec_u8(fprint_len, &|out_buf, out_len| unsafe {
             botan_pubkey_fingerprint(self.obj, hash.as_ptr(), out_buf, out_len)
         })
     }
 
     /// DER encode this public key
     pub fn der_encode(&self) -> Result<Vec<u8>> {
-        let der_len = 4096; // fixme
-        call_botan_ffi_returning_vec_u8(der_len, &|out_buf, out_len| unsafe {
-            botan_pubkey_export(self.obj, out_buf, out_len, 0u32)
-        })
+        #[cfg(feature = "botan3")]
+        {
+            call_botan_ffi_viewing_vec_u8(&|ctx, cb| unsafe {
+                botan_pubkey_view_der(self.obj, ctx, cb)
+            })
+        }
+
+        #[cfg(not(feature = "botan3"))]
+        {
+            let der_len = 4096; // fixme
+            call_botan_ffi_returning_vec_u8(der_len, &|out_buf, out_len| unsafe {
+                botan_pubkey_export(self.obj, out_buf, out_len, 0u32)
+            })
+        }
     }
 
     /// PEM encode this public key
     pub fn pem_encode(&self) -> Result<String> {
-        let pem_len = 4096; // fixme
-        call_botan_ffi_returning_string(pem_len, &|out_buf, out_len| unsafe {
-            botan_pubkey_export(self.obj, out_buf, out_len, 1u32)
+        #[cfg(feature = "botan3")]
+        {
+            call_botan_ffi_viewing_str_fn(&|ctx, cb| unsafe {
+                botan_pubkey_view_pem(self.obj, ctx, cb)
+            })
+        }
+
+        #[cfg(not(feature = "botan3"))]
+        {
+            let pem_len = 4096; // fixme
+            call_botan_ffi_returning_string(pem_len, &|out_buf, out_len| unsafe {
+                botan_pubkey_export(self.obj, out_buf, out_len, 1u32)
+            })
+        }
+    }
+
+    #[cfg(feature = "botan3")]
+    /// Return the encoded elliptic curve point associated with this key
+    ///
+    /// Only valid for EC based keys
+    pub fn ec_public_point(&self) -> Result<Vec<u8>> {
+        call_botan_ffi_viewing_vec_u8(&|ctx, cb| unsafe {
+            botan_pubkey_view_ec_public_point(self.obj, ctx, cb)
         })
     }
 
